#pragma once

#include <Engine/Base/Timer.h>
#include <Engine/Base/CTString.h>
#include <Engine/Base/Lists.h>
#include <Engine/Graphics/DisplayMode.h>
#include <Engine/Math/Functions.h>
#include <Engine/Graphics/OpenGL.h>


#define SQRTTABLESIZE     8192
#define SQRTTABLESIZELOG2 13


// opengl dll settings
#define ODF_ONEWINDOW       (1L<<0)   // supports only one window
#define ODF_FULLSCREENONLY  (1L<<1)   // don't add windowed modes
#define ODF_USEGDIFUNCTIONS (1L<<2)   // don't use wgl functions instead gdi functions

// extension mask
#define GOEXT_MULTITEXTURE        (1UL<<2)    // GL_ARB_multitexture
#define GOEXT_COMPILEDVERTEXARRAY (1UL<<3)    // GL_EXT_compiled_vertex_array
#define GOEXT_SWAPINTERVAL        (1UL<<4)    // WGL_EXT_swap_control
#define GOEXT_EDGECLAMP           (1UL<<5)    // GL_EXT_texture_edge_clamp
#define GOEXT_TEXTURELODBIAS      (1UL<<7)    // GL_EXT_texture_lod_bias
#define GOEXT_TEXTUREANISOTROPY   (1UL<<8)    // GL_EXT_texture_filter_anisotropic
#define GOEXT_VERTEXARRAYRANGE    (1UL<<9)    // GL_NV_vertex_array_range & GL_NV_fence
#define GOEXT_TBUFFER             (1UL<<10)   // 3DFX's T-Buffer (for Motion-blur effect)
#define GOEXT_PNTRIANGLES         (1UL<<11)   // ATI Truform (GL_ATIX_pn_triangles)
#define GOEXT_TEXTURECOMPRESSION  (1UL<<20)   // any texture compression at all?
#define GOEXTC_ARB                (1UL<<21)   // GL_ARB_texture_compression
#define GOEXTC_S3TC               (1UL<<22)   // GL_EXT_texture_compression_s3tc
#define GOEXTC_FXT1               (1UL<<23)   // GL_3DFX_texture_compression_FXT1
#define GOEXTC_LEGACY             (1UL<<24)   // GL_S3_s3tc

// flags
#define GLF_HASACCELERATION  (1UL<<0)   // set if current mode supports hw acceleration
#define GLF_INITONNEXTWINDOW (1UL<<1)   // initialize rendering context on next window
#define GLF_32BITTEXTURES    (1UL<<2)   // set if rendere supports 32-bit textures
#define GLF_SCENEUSESZBUFFER (1UL<<3)   // set if scene renderer should compare trough z-buffer (for second layer)
#define GLF_ADJUSTABLEGAMMA  (1UL<<4)   // set if display allows gamma adjustment


/*
 * OpenGL driver for graphics library.
 */
class COpenGLDriver {
public:
  enum OpenGLDriverType ogd_odtType;    // sub-driver type
  const char *ogd_strName;              // descriptive name
  const char *ogd_strFileName;          // filename of dll to load
  ULONG ogd_ulFlags;
};

/*
 *  Graphics library object.
 */
class ENGINE_API CGfxLibrary {
public:
  // some OpenGL info
  COpenGLDriver go_ogdCurrent;     // current sub-driver
  HINSTANCE     go_hiOpenGLDriver; // opengl dll handle
  HGLRC         go_hglRC;          // opengl rendering context
  HWND          go_hWndLastActive; // last active window
  // driver info constant
  CTString go_strVendor;
  CTString go_strRenderer;
  CTString go_strVersion;
  CTString go_strExtensions;
  CTString go_strSupportedExtensions;
  INDEX go_ctTextureUnits;
  INDEX go_ctDriverChanges;        // count of driver changes
  INDEX go_ctSampleBuffers;        // number of rendering buffers currently on (for T-buffer's motion-blur effect) 0=not supported
  INDEX go_iCurrentWriteBuffer;    // current buffer enabled for writing
  ULONG go_ulSupportedExtensions;  // bit mask for supported extensions
  ULONG go_ulFlags;

  INDEX go_iVSyncsToWait;
  PIX   go_pixMaxTextureDimension;  // maximum allowed texture dimension by OpenGL driver
  FLOAT go_fTextureLODBias;         // current texture lod bias
  FLOAT go_fMaxTextureLODBias;      // absolute maximum possible texture lod bias (0 if biasing is not supported)
  FLOAT go_fMaxTextureAnisotropy;   // maximum possible texture degree of anisotropy (1.0f if isotropic only)
  INDEX go_iMaxTesselationLevel;    // maximum possible truform tesselation level (0=no truform)

  CDisplayMode gl_dmCurrentDisplayMode;  // currently active display mode
  INDEX gl_iFrameNumber;           // frame number currently rendering 
  CTimerValue gl_tvFrameTime;      // time when swapbuffer occured 
  SLONG gl_slAllowedUploadBurst;   // remain upload burst size for this frame (max texture or shadowmap size *2)
  CListHead gl_lhCachedShadows;    // list of all cached shadowmaps
  BOOL gl_bAllowProbing;

  // for profiling
  INDEX gl_ctWorldTriangles;     // how many world polygons were rendered in last frame
  INDEX gl_ctModelTriangles;     // how many model polygons were rendered in last frame
  INDEX gl_ctParticleTriangles;  // how many particle polygons were rendered in last frame
  INDEX gl_ctTotalTriangles;     // total = (world & lensflares) + (model & shadows) + particle + 2D

public:
// implementation:

  BOOL InitSubDriver( COpenGLDriver &ogd);
  void EndSubDriver(  COpenGLDriver &ogd);
  void TestExtension( ULONG ulFlag, const char *strName);
  void AddExtension(  ULONG ulFlag, const char *strName);

  void InitContext(void);
  BOOL SetupPixelFormat( HDC hdc, BOOL bReport=FALSE);
  BOOL CreateContext( HDC hdc);
  BOOL SetCurrentViewport( CViewPort *pvp);

  // for selection
  void UploadPattern( ULONG ulPatternEven, ULONG ulPatternOdd);

  // optimize memory used by cached shadow maps
  void ReduceShadows(void);

  // pretouch all textures and cached shadowmaps
  void PretouchIfNeeded(void);

  /* Construct uninitialized gfx library. */
  CGfxLibrary(void);
	/* Destruct (and clean up). */
  ~CGfxLibrary();

// implementation:
  /* Get list of all supported display modes. */
  CListHead &EnumDisplayModes(void) const;

  /* Start given windowed mode, */
  BOOL StartDisplayMode(PIX pixSizeI, PIX pixSizeJ, enum DisplayDepth dd, enum OpenGLDriverType odt);
  /* Stop windowed mode, */
  void StopDisplayMode(void);

  // swap buffers in a viewport
  void SwapBuffers( CViewPort *pvpToSwap);
  // return current frame number
  inline INDEX GetFrameNumber(void) { return gl_iFrameNumber; };

	/* Lock a raster for drawing. */
  BOOL LockRaster(CRaster *praToLock);
	/* Unlock a raster after drawing. */
  void UnlockRaster(CRaster *praToUnlock);
  /* Lock a drawport for drawing. */
  BOOL LockDrawPort(CDrawPort *pdpToLock);
	/* Unlock a drawport after drawing. */
  void UnlockDrawPort(CDrawPort *pdpToUnlock);

// interface:

	// general functions
	/* Initialize library. */
  void Init(void);
  /* Set if scene renderer for current driver should use allready filled z-buffer. */
  inline void SetSceneRendererUsesZBuffer( BOOL bSceneUsesZBuffer) {
    if( bSceneUsesZBuffer) go_ulFlags |=  GLF_SCENEUSESZBUFFER;
    else                   go_ulFlags &= ~GLF_SCENEUSESZBUFFER; };
  /* Test if scene renderer for current driver should use allready filled z-buffer. */
  inline BOOL IsSceneRendererUsingZBuffer(void) { return (go_ulFlags & GLF_SCENEUSESZBUFFER); };

  // set new custom display mode
  BOOL SetDisplayMode(PIX pixSizeI, PIX pixSizeJ, enum DisplayDepth dd, enum OpenGLDriverType odt);
  // set display mode to original desktop display mode and default ICD driver
  BOOL ResetDisplayMode(void);
  // temporary reset display mode to desktop original mode, but keep driver
  void PauseDisplayMode(void);
  // return to display mode that was active before pausing
  void ResumeDisplayMode(void);
  // get current display mode
  inline void GetDisplayMode( CDisplayMode &dmCurrent) { dmCurrent = gl_dmCurrentDisplayMode; };
  // check if 3d acceleration is on
  inline BOOL IsCurrentModeAccelerated(void) { return (go_ulFlags&GLF_HASACCELERATION); };
  // check if DualHead is on
  inline BOOL IsCurrentModeDH(void) { return gl_dmCurrentDisplayMode.IsDualHead(); };

  // canvas functions
  /* Create a new window canvas. */
  void CreateWindowCanvas( void *hwnd, CViewPort **ppvpNew, CDrawPort **ppdpNewPIX);
  /* Destroy a window canvas. */
  void DestroyWindowCanvas( CViewPort *pvpOld);

  /* Create a work canvas. */
  void CreateWorkCanvas( PIX pixWidth, PIX pixHeight, CDrawPort **ppdpNew);
  /* Destroy a work canvas. */
  void DestroyWorkCanvas( CDrawPort *pdpOld);

  /* Benchmark current opengl driver. */
  void Benchmark( CViewPort *pvp, CDrawPort *pdp);
};


struct MipmapTable {
  INDEX mmt_ctMipmaps;      // number of mip-maps
  SLONG mmt_slTotalSize;    // total size of all mip-maps (in pixels)
  PIX mmt_pixU;             // size of first mip-map
  PIX mmt_pixV;
  SLONG mmt_aslOffsets[MAX_MEX_LOG2]; // offsets of each mip-map (in pixels)
};


// returns number of mip-maps to skip from original texture depending on maximum allowed texture
// size (in pixels!) and maximum allowed texture dimension (in pixels, too!)
extern INDEX ClampTextureSize( PIX pixClampSize, PIX pixClampDimension, PIX pixSizeU, PIX pixSizeV);

// create mip-map table for texture or shadow of given dimensions
extern void MakeMipmapTable( PIX pixU, PIX pixV, MipmapTable &mmt);

// adds 8-bit opaque alpha channel to 24-bit bitmap (in place suported)
extern void AddAlphaChannel( UBYTE *pubSrcBitmap, ULONG *pulDstBitmap, PIX pixSize, UBYTE *pubAlphaBitmap=NULL);
// removes 8-bit alpha channel from 32-bit bitmap (in place suported)
extern void RemoveAlphaChannel( ULONG *pulSrcBitmap, UBYTE *pubDstBitmap, PIX pixSize);

// flips 24 or 32-bit bitmap (iType: 1-horizontal, 2-vertical, 3-diagonal) - in place supported
extern void FlipBitmap( UBYTE *pubSrc, UBYTE *pubDst, PIX pixWidth, PIX pixHeight, INDEX iFlipType, BOOL bAlphaChannel);


// makes ALL lower mipmaps (to size of Nx1 or 1xM) of an specified 32-bit bitmap with eventual filter appliance
// this is done in-place! - all subsequent mip-maps will be posioned just after the original, input bitmap)
// (only first ctFineMips number of mip-maps will be filtered with bilinear subsampling, while
//  all others will be downsampled with nearest-neighbour method with border preservance)
extern void MakeMipmaps( INDEX ctFineMips, ULONG *pulMipmaps, PIX pixWidth, PIX pixHeight, INDEX iFilter);

// colorize mipmaps
extern void ColorizeMipmaps( INDEX i1stMipmapToColorize, ULONG *pulMipmaps, PIX pixWidth, PIX pixHeight);


// performs dithering of a 32-bit bipmap (can be in-place)
extern void DitherBitmap( INDEX iDitherType, ULONG *pulSrc, ULONG *pulDst, PIX pixWidth, PIX pixHeight,
                          PIX pixCanvasWidth=0, PIX pixCanvasHeight=0);
// performs dithering of a 32-bit mipmaps (can be in-place)
extern void DitherMipmaps( INDEX iDitherType, ULONG *pulSrc, ULONG *pulDst, PIX pixWidth, PIX pixHeight);

// adjust bitmap's saturation and/or hue
extern void AdjustBitmapColor( ULONG *pulSrc, ULONG *pulDst, PIX pixWidth, PIX pixHeight, 
                               SLONG const slHueShift, SLONG const slSaturation);

// applies filter to 32-bit bitmap (iFilter from -6 to 6; negative=sharpen, positive=blur, 0=none)
extern void FilterBitmap( INDEX iFilter, ULONG *pulSrc, ULONG *pulDst, PIX pixWidth, PIX pixHeight,
                          PIX pixCanvasWidth=0, PIX pixCanvasHeight=0);


// (re)allocates temporary memory buffer and mark it as in use
extern void *LockTempBuffer( SLONG slRequiredSize, BOOL bForce=FALSE);
// marks temporary memory buffer as unused and eventually frees its memory
extern void UnlockTempBuffer( BOOL bFree=FALSE);
// adjust temporary memory buffer to suitable size
extern void AdjustTempBuffer(void);


// retrives number of mip-maps for the given bitmap size
__forceinline INDEX GetNoOfMipmaps( PIX pixWidth, PIX pixHeight) { return( FastLog2( Min(pixWidth, pixHeight))+1); }

// retrives memory offset of a specified mip-map or a size of all mip-maps (IN PIXELS!)
// (zero offset means first, i.e. largest mip-map)
extern PIX GetMipmapOffset( INDEX iMipLevel, PIX pixWidth, PIX pixHeight);

// return offset, pointer and dimensions of mipmap of specified size inside texture or shadowmap mipmaps
extern INDEX GetMipmapOfSize( PIX pixSize, ULONG *&pulFrame, PIX &pixProbeWidth, PIX &pixProbeHeight);

// texture parameters for opengl texture state changes
class CTexParams {
public:
  INDEX  tp_iFilter;            // OpenGL texture mapping mode
  FLOAT  tp_fAnisotropy;        // texture degree of anisotropy (>=1.0f; 1.0=isotropic, default)
  BOOL   tp_bSingleMipmap;      // texture has only one mipmap
  GLuint tp_eWrapU, tp_eWrapV;  // wrapping states
  inline CTexParams(void) { Clear(); tp_bSingleMipmap = FALSE; };
  inline void Clear(void) { tp_iFilter = 00; tp_fAnisotropy = 0.0f; tp_eWrapU = tp_eWrapV = GL_FALSE; };
  inline BOOL IsEqual( CTexParams tp) { return tp_iFilter==tp.tp_iFilter && tp_fAnisotropy==tp_fAnisotropy && 
                                               tp_eWrapU==tp.tp_eWrapU && tp_eWrapV==tp.tp_eWrapV; };
};

// global texture parameters in engine
ENGINE_API extern CTexParams _tpGlobal;
inline void SetTextureClampingMode( GLuint eWrapU, GLuint eWrapV) {
  _tpGlobal.tp_eWrapU = eWrapU;
  _tpGlobal.tp_eWrapV = eWrapV;
};

// pointer to global graphics library object
ENGINE_API extern CGfxLibrary *_pGfx;
// forced texture upload quality (0 = default, 16 = force 16-bit, 32 = force 32-bit)
ENGINE_API extern INDEX _iTexForcedQuality;
